/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet host watchdog
 *
 * Chip type           : Atmega168 or Atmega328, Atmega644 with ENC28J60
 * Note: there is a version number in the text. Search for tuxgraphics
 *
 * snmpwalk  -c public -v 1  10.0.0.29 1.3.6.1.4.1.42
 * TUXGRAPHICS-HWD-MIB::name = STRING: host watchdog
 * TUXGRAPHICS-HWD-MIB::resetCnt = INTEGER: 0
 * TUXGRAPHICS-HWD-MIB::status = INTEGER: 0
 * TUXGRAPHICS-HWD-MIB::state = STRING: active
 * TUXGRAPHICS-HWD-MIB::voltage0 = STRING: 4.44V
 * TUXGRAPHICS-HWD-MIB::intvoltage0 = INTEGER: 444
 * TUXGRAPHICS-HWD-MIB::voltage1 = STRING: 4.86V
 * TUXGRAPHICS-HWD-MIB::intvoltage1 = INTEGER: 486
 * End of MIB
 *
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"
#include "analog.h"
#include "websrv_help_functions.h"

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x11,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,29};
//static uint8_t myip[4] = {192,168,0,82};
static uint8_t monitoredhost[4] = {10,0,0,5};
//static uint8_t monitoredhost[4] = {192,168,0,1};
// the password string (only a-z,0-9,_ characters):
static char password[9]="secret"; // must not be longer than 8 char
// you can calibrate the board's voltage measurement function
// a bit by changing the following two factors. 
// This may be needed because the internal reference 
// voltage of 1.1V can be a bit off.
// As per atmega datasheet it can vary up to 10% (1V to 1.2V)
// from chip to chip. 
// That is: different chips have different AREF voltages but
// the value for a specific atmega chip stays the same and
// does not fluctuate. Cange only one of the values ADCDIV_VAL or
// ADCMUL_VAL. Increasing ADCMUL_VAL gives higher voltages and lowering
// it gives smaller voltages.
#define ADCMUL_VAL 308
#define ADCDIV_VAL 409
//
static uint8_t gwip[4]; // optional gwip
// listen port for tcp/www:
static uint16_t mywwwport=80;
//static uint16_t mywwwport=82;
static char hwd_name[25] = "host watchdog"; // used for tuxgraphicsHWD.name
// the community string is like a password:
static char my_snmp_community[10] = "public";
// our OIDs have the format 1.3.6.1.4.1.ENTERPRISE_OID_NUM.X 
#define ENTERPRISE_OID_NUM 42
//
//------------------- do not change below this line
// our OIDs have the format 1.3.6.1.4.1.42.X with X from 0 to MAX_OID_LAST_DIGIT_X:
#define MAX_OID_LAST_DIGIT_X 7
static volatile uint8_t hoststatus=1; // 0=all ping/replies OK, otherwise amount of missed counts, used for tuxgraphicsHWD.status
static uint8_t resetcount=0; // how often reset was executed, used for tuxgraphicsHWD.resetCnt
static uint8_t hostbackafterreboot=0; // wait for on answer until we really activate the wd
static uint8_t sendping=1; // 1 we send ping (and receive ping), 0 we receive ping only
static volatile uint8_t pingnow=1; // 1 means time has run out send a ping now
static volatile uint8_t resetnow=0; 
static uint8_t reinitmac=0; 
static uint8_t do_extra_arp=0; 
static uint8_t pinginterval=10; // after how many seconds to send or receive a ping (value range: 2 - 250)
static volatile uint8_t pingtimer=1; // > 0 means wd running
//
#define BUFFER_SIZE 610
static uint8_t buf[BUFFER_SIZE+1];
static uint16_t gPlen;
// global string buffer
#define STR_BUFFER_SIZE 24
static char gStrbuf[STR_BUFFER_SIZE+1];
// we have two ways to read the voltages (as integers and as strings)
// if the voltage is moving a lot then those two could be different even
// for the same adc because they are read at different times. That does
// not look nice so we read them centrally and use only the pre-stored values.
static volatile uint8_t measurement_timer=0;
static uint16_t volt_times_100[2]={0,0};
//
// 
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}

uint16_t print_line_mobile_webpage(uint16_t plen)
{
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        return(plen);
}

uint16_t print_webpage_cf(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> <a href=http://tuxgraphics.org/hwd4.html>[help]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>WD config</h2>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre><form action=/mm method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR(" monitored IP: <input type=text name=ip value="));
        mk_net_str(gStrbuf,monitoredhost,4,'.',10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(">\n      Gateway: <input type=text name=gp value="));
        mk_net_str(gStrbuf,gwip,4,'.',10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(">\nping interval: <input type=text name=pi value="));
        itoa(pinginterval,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("> <input type=checkbox name=sp value=1 "));
        if (sendping){
                plen=fill_tcp_data_p(buf,plen,PSTR("checked"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR(">Send ping\n     password: <input type=password name=pw>\n<input type=submit value=change></form>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<hr>"));

        return(plen);
}

// prepare the main webpage by writing the data to the tcp send buffer
uint16_t print_webpage_main(uint8_t *buf)
{
        uint16_t plen;
        // periodic refresh every 60sec:
        //plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\nRefresh: 60\r\n\r\n"));
        // no periodic refresh:
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/cf>[cfg]</a> <a href=/vv>[voltages]</a> <a href=/ac>[actions]</a> <a href=http://tuxgraphics.org/hwd4.html>[help]</a> <a href=/>[refresh]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<h2>"));
        plen=fill_tcp_data(buf,plen,hwd_name);
        plen=fill_tcp_data_p(buf,plen,PSTR("</h2>\n<pre>\nWD Status: "));
        // if watchdog is off then status is unknown:
        if (pingtimer==0){
                plen=fill_tcp_data_p(buf,plen,PSTR("-"));
        }else{
                if (hoststatus==0){
                        plen=fill_tcp_data_p(buf,plen,PSTR("<font color=#00ff00>OK</font>"));
                }else{
                        itoa(hoststatus,gStrbuf,10);
                        plen=fill_tcp_data(buf,plen,gStrbuf);
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR(" [reset cnt:"));
        itoa(resetcount,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(", state: "));
        // will also be available in tuxgraphicsHWD.state
        if (pingtimer==0){
                plen=fill_tcp_data_p(buf,plen,PSTR("stopped"));
        }else{
                if (hostbackafterreboot){
                        plen=fill_tcp_data_p(buf,plen,PSTR("active"));
                }else{
                        plen=fill_tcp_data_p(buf,plen,PSTR("passive"));
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("]\n\n monitored IP: "));
        mk_net_str(gStrbuf,monitoredhost,4,'.',10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("\n        GW IP: "));
        mk_net_str(gStrbuf,gwip,4,'.',10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("\nping interval: "));
        itoa(pinginterval,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(" sec\n    send ping: "));
        if (sendping){
                plen=fill_tcp_data_p(buf,plen,PSTR("yes"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("no, expect pings"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\n</pre>\n<hr><p align=right>4.2,tuxgraphics</p>"));

        return(plen);
}

// convert an integer to a string and
// insert a '.' before the second to the last digit/character.
// "100" will become "1.00"
// This function writes to result_str_buf
// The user is responsible for allocating enough space.
void adddecimalpoint(int16_t num,char *result_str_buf){
        uint8_t i;
        itoa(num/100,result_str_buf,10);  // digits before decimal point
        i=strlen(result_str_buf);
        result_str_buf[i]='.';
        // digits after decimal point:
        if (abs(num)%100<10){
                result_str_buf[i+1]='0';
                i++;
        }
        itoa(abs(num)%100,&(result_str_buf[i+1]),10); 
}

// multiply a large number.
// We need this function because the avr-gcc 32bit datatypes don't work. 
// The 24 bit number is represented in lnum[6] with each component
// being 4 bit wide. We need such small numbers (and not 8bit) because
// "lnum[i]*snum + tmp" must never be bigger than 16bit.
// The 24 bit number is 6 times 4 bit: lnum[6]
// lnum[0] is the least significat 4 bits.
// The result is returned to lnum;
void mult24(uint8_t *lnum,uint16_t snum)
{
	uint8_t i=0;
	uint16_t tmp=0; // 16bit
	// start with lowest byte:
	while(i<6){
		tmp=lnum[i]*snum + tmp;
		lnum[i]=tmp & 0xf;
		tmp=tmp>>4;
		i++;
	}
}

// devide the 24 bit integer unsigned char i[6] by a 16bit number (snum) 
// and return the value in i[6] 
// i[0] is the least significat 4 bits.
// remain is just a simple 16 bit variable
// We must use 4 bit wide numbers to avoid that the remainder shifted by
// the number width + i[loop] gets bigger than 16bit
void div24(uint8_t *i,uint16_t *remain,uint16_t snum){
        int8_t loop;
        uint16_t tmp; // 16 bit
        // start with highest half byte (nibble)
        loop=5;
        *remain=0; // 16 bit
        while(loop>=0){
                tmp=(*remain <<4) + i[loop];
                i[loop]=(tmp / snum)&0xf;
                *remain=tmp % snum;
                loop--;
        }
}

// analog voltage measurement:
uint16_t adc_to_volttimes100(uint8_t adc_num){
        uint16_t tmp;
        uint8_t j;
        uint8_t num24bit[6]; // 24 bit number split in nibbles
        //
        // 2^16 = 65536 max range for uint16_t, we can not multiply
        // the convertanalog result with more than 16 otherwise the
        // datatype will overflow
        // 1.1V * (270+10)/10 * adcreading / 4096 = Vin before resistor
        // The max adc value is actually 4092=16*1023/4
        tmp=getanalogresult(adc_num);
        // debug:
        //itoa(tmp,result_str_buf,10); 
        //return;
        // convert to num24bit:
        j=0;
        while(j<6){
                num24bit[j]=tmp & 0xf;
                tmp=tmp>>4;
                j++;
        }
        mult24(num24bit,ADCMUL_VAL); 
        div24(num24bit,&tmp,ADCDIV_VAL);
        // round if there is a big enough remainder:
        if (tmp>ADCDIV_VAL/2){ num24bit[0]++; }
        j=0;
        tmp=0;
        // convert num24bit to 16bit int:
        while(j<6){
                tmp|=num24bit[j]<<(4*j);
                j++;
        }
        return(tmp);
}

void write_voltage_string(uint16_t volttimes100,char *result_str_buf){
        adddecimalpoint(volttimes100,result_str_buf); // give V*100 to adddecimalpoint
        strcat(result_str_buf,"V");
}

uint16_t print_webpage_voltages(uint8_t *buf)
{
        uint16_t plen;
        int8_t i;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/vv>[refresh]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Voltages</h2>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>\n"));
        i=0;
        while(i<2){ // ADC0 to ADC1
                plen=fill_tcp_data_p(buf,plen,PSTR("adc"));
                itoa(i,gStrbuf,10); // convert integer to string
                plen=fill_tcp_data(buf,plen,gStrbuf); // switch number in ascii
                plen=fill_tcp_data_p(buf,plen,PSTR(": "));
                write_voltage_string(adc_to_volttimes100(i),gStrbuf);
                plen=fill_tcp_data(buf,plen,gStrbuf); 
                plen=fill_tcp_data(buf,plen,"\n"); 
                i++;
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<pre><hr>\n"));
        return(plen);
}

uint16_t print_webpage_now(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> <a href=http://tuxgraphics.org/hwd4.html>[help]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Actions</h2>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/ack method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("Passw: <input type=password size=8 name=pw><input type=hidden name=rst value=1><input type=submit value=\"reboot host now\"></form>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<hr>\n<form action=/ack method=get>"));
        if (pingtimer>0){
                plen=fill_tcp_data_p(buf,plen,PSTR("Passw: <input type=password size=8 name=pw><input type=hidden name=stp value=1><input type=submit value=\"stop watchdog now\"></form><hr>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("Passw: <input type=password size=8 name=pw><input type=hidden name=srt value=1><input type=submit value=\"start watchdog now\"></form>\n</pre><hr>"));
        }
        return(plen);
}

uint16_t print_webpage_confirm(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK</h2> <a href=/>-&gt;continue</a>\n"));
        return(plen);
}

void resethost(void)
{
        PORTD|= (1<<PIND7);// transistor on
        resetcount++;
        hostbackafterreboot=0;
        resetnow=1; // trigger counter to clear transistor 2 sec later
}

void store_in_eeprom(void)
{
        eeprom_write_byte((uint8_t *)20,19); // magic number
        eeprom_write_block((uint8_t *)gwip,(void *)21,sizeof(gwip));
        eeprom_write_block((uint8_t *)monitoredhost,(void *)26,sizeof(monitoredhost));
        eeprom_write_byte((uint8_t *)31,pinginterval); 
        eeprom_write_byte((uint8_t *)32,sendping); 
        reinitmac=1;
        hostbackafterreboot=0;
        resetcount=0;
        hoststatus=1;
}

// takes a string of the form ack?pw=xxx&rst=1 and analyse it
// the first "/" is already removed
// return values:  0 error: 404 Not Found
//                -1 error: 401 Unauthorized
//                 1 main page
int8_t analyse_get_url(char *str)
{
        static uint8_t tmpip[4];
        if (str[0]==' '){
                return(1); // end of url, main page
        }
        if (strncmp("vv ",str,3)==0){
                gPlen=print_webpage_voltages(buf);
                return(10);
        }
        if (strncmp("ac ",str,3)==0){
                gPlen=print_webpage_now(buf);
                return(10);
        }
        if (strncmp("favicon.ico",str,11)==0){
                gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("http://tuxgraphics.org/ico/wd.ico"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("\r\n\r\nContent-Type: text/html\r\n\r\n"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>301 Moved Permanently</h1>\n"));
                return(10);
        }
        // --------
        if (strncmp("cf ",str,3)==0){
                gPlen=print_webpage_cf(buf);
                return(10);
        }
        // actions: reboot stop start
        if (strncmp("ack",str,3)==0){
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"rst")){
                                        resethost();
                                        gPlen=print_webpage_confirm(buf);
                                        return(10);
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"stp")){
                                        pingtimer=0; // wd off 
                                        LEDOFF;
                                        gPlen=print_webpage_confirm(buf);
                                        return(10);
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"srt")){
                                        pingtimer=1; // wd on 
                                        hoststatus=1;
                                        if (sendping && pinginterval>5){
                                                pingtimer=pinginterval-4;
                                        }
                                        gPlen=print_webpage_confirm(buf);
                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        // modify settings:
        if (strncmp("mm",str,2)==0){
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"ip")){
                                        urldecode(gStrbuf);
                                        if (parse_ip(tmpip,gStrbuf)){
                                                return(0); // not a good IP
                                        }
                                        // good ip:
                                        monitoredhost[0]=tmpip[0];
                                        monitoredhost[1]=tmpip[1];
                                        monitoredhost[2]=tmpip[2];
                                        monitoredhost[3]=tmpip[3];
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"gp")){
                                        urldecode(gStrbuf);
                                        if (parse_ip(tmpip,gStrbuf)){
                                                // not a good IP, do not use GW:
                                                gwip[0]=0;
                                                gwip[1]=0;
                                                gwip[2]=0;
                                                gwip[3]=0;

                                        }
                                        // good ip:
                                        gwip[0]=tmpip[0];
                                        gwip[1]=tmpip[1];
                                        gwip[2]=tmpip[2];
                                        gwip[3]=tmpip[3];
                                }
                                // some default values:
                                pinginterval=5;
                                sendping=0;
                                //
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"sp")){
                                        sendping=1;
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pi")){
                                        urldecode(gStrbuf);
                                        pinginterval=atoi(gStrbuf);
                                        if (pinginterval<2){
                                                pinginterval=2;
                                        }
                                        if (pinginterval>250){
                                                pinginterval=250;
                                        }
                                        // accellerate the status update:
                                        if (sendping && pinginterval>5){
                                                pingtimer=pinginterval-4;
                                        }
                                }
                                store_in_eeprom();
                                gPlen=print_webpage_confirm(buf);
                                return(10);
                        }
                }
                return(-1);
        }
        return(0);
}

void step_hoststatus(void)
{
        hoststatus++; // get's cleared as soon as we get a ping
        // refresh arp in case mac has changed:
        if (sendping==1 && hoststatus>1){
                do_extra_arp=1;
        }
        if (hoststatus==0xff){
                hoststatus=1; // wrap to 1, not zero
        }
        // LEDON at first failure:
        //LEDON;
}

// timer interrupt, step counter every second
ISR(TIMER1_COMPA_vect){
        measurement_timer++;
        //-----
        if (resetnow>2){
                PORTD &= ~(1<<PIND7);// transistor off
                resetnow=0;
        }
        if (resetnow){
                resetnow++;
        }
        if (hoststatus>5 && hostbackafterreboot){
                resethost();
        }
        //-----
        if (pingtimer>pinginterval){
                // do not wrap to 0 continue with 1:
                pingtimer=1;
        }
        //----- set a flag to send a ping as soon as possible 
        // this is to avoid interference with web traffic
        if (pingtimer == pinginterval){
                pingnow=1;
        }
        // pingtimer==0 means off
        if (pingtimer>0){
                if (sendping==0 && pingnow){
                        // we should receive a ping
                        pingnow=0;
                        step_hoststatus();
                }
                pingtimer++;
        }
        // for security reasons as we increas pingtimer also below in main
        if (pingtimer>252){
                pingtimer=1;
        }
}


// Generate an interrup about ever 1s form the 12.5MHz system clock
// Since we have that 1024 prescaler we do not really generate a second
// (1.00000256000655361677s) 
void timer_init(void)
{
        /* write high byte first for 16 bit register access: */
        TCNT1H=0;  /* set counter to zero*/
        TCNT1L=0;
        // Mode 4 table 14-4 page 132. CTC mode and top in OCR1A
        // WGM13=0, WGM12=1, WGM11=0, WGM10=0
        TCCR1A=(0<<COM1B1)|(0<<COM1B0)|(0<<WGM11);
        TCCR1B=(1<<CS12)|(1<<CS10)|(1<<WGM12)|(0<<WGM13); // crystal clock/1024

        // At what value to cause interrupt. You can use this for calibration
        // of the clock. Theoretical value for 12.5MHz: 12207=0x2f and 0xaf
        OCR1AH=0x2f;
        OCR1AL=0xaf;
        // interrupt mask bit:
        TIMSK1 = (1 << OCIE1A);
}

void ping_callback(uint8_t *from_ip){
        uint8_t i=0;
        while(i<4){
                if(from_ip[i]!=monitoredhost[i]){
                        // not the host we wanted to be pinged from:
                        return;
                }
                i++;
        }
        hoststatus=0;
        hostbackafterreboot=1;
        // LEDOFF as soon as host is back and we got pinged
        //LEDOFF;
}

// functions for snmp -- simple network management protocol

// asn1 encoded messages are really an awful security risk because
// of all the length fields. Any of them could be bogus. Do a basic
// bouds check  of important fields that we will use later on
// return 0 on failure
uint8_t snmp_sanitycheck(uint8_t  *d,uint8_t snmplen)
{
        uint8_t i;
        if (*(d+1)!=snmplen-2){
                return(0);
        }
        // community srting len:
	i=*(d+6);
        if (i>snmplen-4){
                return(0);
        }
        // request_type:
        i=*(d+6+i+1);
        // request_type=
        // 0xa0 = get
        // 0xa1 = get next
        // 0xa3 = set
        // 0xa2 = response
	if (i<0xa0 || i>0xa3){
		return(0); // not a valid request type or invalid community string length
	}
        // request ID len:
	i=*(d+6+(*(d+6))+1+3);
	if (i<1 || i>4){	
		// error in message. max should be 32bit integer, min 8 bit
		return(0);
	}
        return(1);
}

// the community string is like a "password". Only if it matches
// we will answer with data. The community "public" is used frequently
// The snmp agent (this code) should not answer at all if there is
// a mismatch.
uint8_t snmp_check_community(uint8_t *d,char *community_to_match)
{
// example: 0x30,0x26,0x02,0x01,0x00,0x04,0x06,0x70,0x75,0x62,0x6c,0x69,0x63
//                              |    |    |     |
//                              ver  str  6byte The-string
// first octet must be 0x30 (sequence)
// length must be bigger than 0x20
// we support version 1 of snmp: 0x02, 0x01, 0xVER 
// match community: 0x04, 0xlen, community string
	if (*d!=0x30){
		return(0); // error
	}
	if (*(d+4)>0){ // snmp version, 0=version 1
		return(0); // error
	}
	if (*(d+5)!=0x04){ // must be of type string
		return(0); // error
	}
	if (strlen(community_to_match)!=*(d+6)){
		return(0); // error
	}
	if (strncmp(community_to_match,(char *)(d+7),*(d+6))!=0){
		return(0); // error
	}
	return(1);
}

// Format of messages: tag len value
// The tag is one of the following ASN.1 DER data types
//  0x30 sequence
//  0x02 integer
//  0x04 byte string (octetstring)
//  0x05 null
//  0x06 OID  (the snmp OID is always terminated with NULL: 0x05 0x00)
// see http://www.faqs.org/rfcs/rfc1228.html , http://www.ietf.org/rfc/rfc1028.txt
// http://www.ietf.org/rfc/rfc1157.txt , 

//--  The hierarchy is:
//--  iso             OBJECT IDENTIFIER ::= { 1 }
//--  org             OBJECT IDENTIFIER ::= { iso 3 }
//--  dod             OBJECT IDENTIFIER ::= { org 6 }
//--  internet        OBJECT IDENTIFIER ::= { dod 1 }
//--  private         OBJECT IDENTIFIER ::= { internet 4 }
//--  enterprises     OBJECT IDENTIFIER ::= { private 1 }
//--  thisOne        OBJECT IDENTIFIER ::= { enterprises ENTERPRISE_OID_NUM }
//--  hwdmib         OBJECT IDENTIFIER ::= { thisOne 1 } 1.3.6.1.4.1.ENTERPRISE_OID_NUM.1
// We did not register the "thisOne" enterprise. We just use it
// It is unlikely that it will overlap
// The obove sequence is encoded as 2B 06 01  04 01 2A 01

// check if it is get or not 
// find the start of the seq tag containing the seq of OIDs
// return values: offset_to_start_of_getSeq return of function: 0=no get no getnext, 1 = get, 2 = getnext
// offset_to_start_of_getSeq is a length offset
uint8_t snmp_check_get_or_getnext(uint8_t  *d,uint8_t *offset_to_start_of_getSeq)
{
	uint8_t community_str_len;
	uint8_t requIDlen;
	uint8_t reqtype;
	community_str_len=*(d+6);
	// must be 0xa0 for a get-request and 0xa1 for getnext:
        reqtype=*(d+6+community_str_len+1);
	if (reqtype!=0xa0 && reqtype!=0xa1){
		return(0); // not a get and not getnext
	}
	// the request ID is an integer and can be 1 byte or more
	requIDlen=*(d+6+community_str_len+1+3);
	if (*(d+6+community_str_len+1+3+requIDlen+7)!=0x30){
		// must be seq tag
		return(0);
	}
	*offset_to_start_of_getSeq=(6+community_str_len+1+3+requIDlen+7);
        if (reqtype==0xa0){
                return(1);
        }
        return(2);
}

// 0x30,0x1c,0x30,0x0c,...,0x30,0x0c,...
// d must point to the first 0x30
// skip over X amount of seq in the seq tag
uint8_t *skipseq(uint8_t *d,uint8_t no_of_seq_to_skip){
	uint8_t skiplen=2;
	while(no_of_seq_to_skip){
		skiplen=skiplen+*(d+skiplen+1)+2;
		no_of_seq_to_skip--;
	}
	return(d+skiplen);
}

// 0x30,0x1c,0x30,0x0c,...,0x30,0x0c,...
// d must point to the first 0x30
// return amount of SEQ in this SEQ
// The structure of SNMP get is like this:
//
//      SEQUENCE {
//         SEQUENCE {
//            OBJECTIDENTIFIER 1.3.6.1.....
//            NULL 
//         }
//         SEQUENCE {
//            OBJECTIDENTIFIER 1.3.6.1.....
//            NULL 
//         }
//      }
// There might be 1 or more inner sequences
// This function finds out how many sequences there are inside
uint8_t snmp_numofSeq(uint8_t *d){
	uint8_t outerlen;
	uint8_t seqcnt=0;
	uint8_t this_inner_len;
	uint8_t *dptr;
	if (*d!=0x30){
		return(0); // error
	}
	outerlen=*(d+1);
	this_inner_len=0;
	while(outerlen-this_inner_len>2){
		outerlen=outerlen-2-this_inner_len;
		dptr=skipseq(d,seqcnt);
		this_inner_len=*(dptr+1);
		seqcnt++;
	}
	return(seqcnt);
}
	
// the oid is just the numbers after 1.3.6.1
// seq_str_ptr must point to the outer seq (0x30...) enclosing all the OID value pairs
// itemnumer=0 get the first OID
// itemnumer=1 get the second OID (works only if numofSeq>=1)
// returns the length of the oid string (the remains after 1.3.6.1)
// The array oid can take max length of 8;
uint8_t snmp_get_oid_string(uint8_t *seq_str_ptr,uint8_t itemnumer,uint8_t *oid)
{
	uint8_t *innerseq_ptr;
	uint8_t len,i,rval;
	innerseq_ptr=skipseq(seq_str_ptr,itemnumer);
	len=*(innerseq_ptr+3);
	if (*(innerseq_ptr+5)!=0x06){ // 1.3.6
		return(0);
	}
	if (*(innerseq_ptr+6)!=0x01){ // 1.3.6.1
		return(0);
	}
	i=0;
	len-=3;
	rval=len;
	if (rval>7){
		return(0);
	}
	while(len){
		*(oid+i)=*(innerseq_ptr+7+i);
		len--;
		i++;
	}
	return(rval);
}

//0x2b,0x06,0x01,0x04,0x01,0x2a,0x01,  = 1.3.6.1.4.1.ENTERPRISE_OID_NUM.1
// return -1 if oid is not in our tree or after it
// return the last digit of the oid that must be answered with if it is in
// our tree. This assumes that all oids are of the format myoid.X and X goes
// from 0..X with all X beeing consecutive numbers. No check is done
// here for a max value of X.
// We use here short OIDs that is: the oid variable contains only the
// digits after 1.3.6.1
int8_t snmp_find_next_oid(uint8_t *oid,uint8_t oidlen)
{
        uint8_t myoid[4]={4,1,ENTERPRISE_OID_NUM}; 
#define MYOID_LEN 3
	int8_t i=0;
	while(i<MYOID_LEN && i < oidlen){
		if (oid[i]>myoid[i]){
			// not in our tree
			return(-1);
		}
		if (oid[i]<myoid[i]){
			// it's way before, return the first element
			return(0);
		}
		i++;
	}
	// oid and myoid are now the same, oid may have more or less digits
	if (oidlen<=MYOID_LEN){
		// return the first element
		return(0);
	}
        // return the one after:
        return(oid[MYOID_LEN]+1);
}

// seq_str_ptr must point to the outer seq (0x30...) enclosing all the OID value pairs
// returns 0 if OK or the index to the inner seq which is after our tree 
// and must this be answered with an error 
// E.g return 1 if the first oid is not before our OIDs
// All our OIDs have the format 1.3.6.1.4.1.ENTERPRISE_OID_NUM.X  and X is 0..MAX_OID_LAST_DIGIT_X
// The X is then filled into last_digits_of_oids[i]
uint8_t snmp_find_oids_from_getnext(uint8_t *seq_str_ptr,uint8_t max_num_of_oids_available,uint8_t *last_digits_of_oids)
{
        uint8_t i=0;
        int8_t oid_last_digit;
        uint8_t oidlen;
        uint8_t oid[8];
        while(max_num_of_oids_available){
                max_num_of_oids_available--;
                // the first oid in snmp_get_oid_string is i=0
                oidlen=snmp_get_oid_string(seq_str_ptr,i,oid);
                oid_last_digit=snmp_find_next_oid(oid,oidlen);
                last_digits_of_oids[i]=oid_last_digit;
                i++; 
                if (oid_last_digit<0){
                        // error not a oid in this tree
                        return(i);
                }
                if (oid_last_digit>MAX_OID_LAST_DIGIT_X){
                        return(i);
                }
        }
        return(0);
}

// seq_str_ptr must point to the outer seq (0x30...) enclosing all the OID value pairs
// returns 0 if OK or the index to the inner seq which is wrong.
// E.g 1 if the first oid is not one of our OIDs.
// All our OIDs have the format 1.3.6.1.4.1.ENTERPRISE_OID_NUM.X  and X is 0..MAX_OID_LAST_DIGIT_X
// The X is then filled into last_digits_of_oids[i]
uint8_t snmp_ckeck_if_all_are_my_oids(uint8_t *seq_str_ptr,uint8_t max_num_of_oids_available,uint8_t *last_digits_of_oids)
{
        uint8_t i=0;
        uint8_t oidlen;
        uint8_t oid[8];
        while(max_num_of_oids_available){
                max_num_of_oids_available--;
                // the first oid in snmp_get_oid_string is i=0
                oidlen=snmp_get_oid_string(seq_str_ptr,i,oid);
                last_digits_of_oids[i]=oid[oidlen-1];
                i++; 
                if (oidlen!=4){
                        return(i);
                }
                // my oid=4.1.ENTERPRISE_OID_NUM.X
                if (oid[0]!=4||oid[1]!=1||oid[2]!=ENTERPRISE_OID_NUM){
                        return(i);
                }
                if (oid[3]>MAX_OID_LAST_DIGIT_X){
                        return(i);
                }
        }
        return(0);
}

// make an error response from a get, get next, set
//ErrorStatus ::= INTEGER {
//                              noError(0),
//                              tooBig(1),
//                              noSuchName(2),
//                              badValue(3),
//                              readOnly(4)
//                              genErr(5)
//                          }
// This function modifies the data that d points to.
// errorindex is normally 1 (if there was only one oid in the get).
// d must point to the start of the snmp part of the udp message (=udp data part)
void snmp_make_error_response(uint8_t *d,uint8_t errorcode,uint8_t errorindex)
{
	uint8_t community_str_len;
	uint8_t requIDlen;
	community_str_len=*(d+6);
	// request_type must be 0xa2 for a response:
	*(d+6+community_str_len+1)=0xa2;
	// the request ID is an integer and can be 1 byte or more
	requIDlen=*(d+6+community_str_len+1+3);
	// write error status and error index:
	*(d+6+community_str_len+1+3+requIDlen+3)=errorcode;
	*(d+6+community_str_len+1+3+requIDlen+3+3)=errorindex;
}

// write a seq with OID value pair starting at position that d points to.
// 0x30 (seq tag) is the first byte that will be written. The oid variable
// is an array containing the oid numbers after 1.3.6.1 (=2b 06 01)
// oidlen is the number of bytes after (1.3.6.1, whatever is in *oid).
//         SEQUENCE {
//            OBJECTIDENTIFIER 1.3.6.1.....
//            Value
//         }
// The data type (valuetype) of value is usually one of:
//  0x02 integer
//  0x04 byte string (octetstring)
//
//  vallen is the number of bytes that *val is long. (1 for uint8_t, strlen in case of a string)
//  The function returns the amount of bytes written
uint8_t snmp_fill_oid_val_pair(uint8_t *d,uint8_t *oid, uint8_t oidlen,uint8_t valuetype,uint8_t *val, uint8_t vallen)
{
        uint8_t bytecnt;
        uint8_t i=0;
        d[0]=0x30;
        // next byte is len and will be written later
        d[2]=0x6; // oid tag
        d[3]=oidlen+3;
        d[4]=0x2b; d[5]=0x6; d[6]=0x1;
        while(oidlen){
                d[7+i]=oid[i];
                i++;
                oidlen--;
        }
        d[7+i]=valuetype;
        d[8+i]=vallen;
        bytecnt=9+i;
        i=0;
        while(vallen){
                d[bytecnt+i]=val[i];
                vallen--;
                i++;
        }
        // len field:
        d[1]=bytecnt+i-2;
        return(bytecnt+i);
}

uint16_t byteswap(uint16_t i){
        uint8_t tmp;
        // We need to byteswap the integer such
        // that it is in network order:
        tmp=i&0xff;
        i=i>>8;
        i=i|(tmp<<8);
        return(i);
}

// fill the oid and data part of an snmp answer and update
// the length fields
// snmpdat pointer in buf to the start of the udp snmp data
// offset_to_seq pointer to the seq tag that encloses all the seq/oid tags
// last_digits_of_oids array with OIDs to answer for
// snmp_num_of_oids length of last_digits_of_oids
// return snmpdatalen (amount of udp data bytes)
uint8_t snmp_fill_answer_with_data(uint8_t *snmpdat,uint8_t offset_to_seq,uint8_t *last_digits_of_oids,uint8_t snmp_num_of_oids)
{
        uint8_t myoid[4]={4,1,ENTERPRISE_OID_NUM,0}; // the last digit will differ and be replaced by one of the digits stored in last_digits_of_oids
        uint8_t community_str_len;
	uint8_t requIDlen;
        uint16_t volt_net_order;
        uint8_t totlen_i_seq;
        int8_t j;
        uint8_t snmpdatalen;
        // loop now over all the OIDs:
        j=0;totlen_i_seq=0;
        while (j<snmp_num_of_oids){
                myoid[3]=last_digits_of_oids[j];
                switch (last_digits_of_oids[j]) {
                        case 0:
                        // tuxgraphicsHWD.name=1.3.6.1.4.1.ENTERPRISE_OID_NUM.0
                        totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x4,(uint8_t *)hwd_name,strlen(hwd_name));
                        //
                        break;
                        case 1:
                                // tuxgraphicsHWD.resetCnt=1.3.6.1.4.1.ENTERPRISE_OID_NUM.1
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x2,(uint8_t *)&resetcount,1);
                                break;
                        case 2:
                                // tuxgraphicsHWD.status=1.3.6.1.4.1.ENTERPRISE_OID_NUM.2
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x2,(uint8_t *)&hoststatus,1);
                                break;
                        case 3:
                                // tuxgraphicsHWD.state=1.3.6.1.4.1.ENTERPRISE_OID_NUM.3
                                if (pingtimer==0){
                                        strcpy(gStrbuf,"stopped");
                                }else{
                                        if (hostbackafterreboot){
                                                strcpy(gStrbuf,"active");
                                        }else{
                                                strcpy(gStrbuf,"passive");
                                        }
                                }
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x4,(uint8_t *)gStrbuf,strlen(gStrbuf));
                                break;
                        case 4:
                                // tuxgraphicsHWD.voltage0=1.3.6.1.4.1.ENTERPRISE_OID_NUM.4
                                write_voltage_string(volt_times_100[0],gStrbuf);
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x4,(uint8_t *)gStrbuf,strlen(gStrbuf));
                                break;
                        case 5:
                                // tuxgraphicsHWD.intvoltage0=1.3.6.1.4.1.ENTERPRISE_OID_NUM.5
                                // We need to byteswap the integer such
                                // that it is in network order:
                                volt_net_order=byteswap(volt_times_100[0]);
                                // send the 2 byte long number:
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x2,(uint8_t *)&volt_net_order,2); // we are passing 2 times 8 bit in volt_times_100
                                break;
                        case 6:
                                // tuxgraphicsHWD.voltage0=1.3.6.1.4.1.ENTERPRISE_OID_NUM.6
                                write_voltage_string(volt_times_100[1],gStrbuf);
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x4,(uint8_t *)gStrbuf,strlen(gStrbuf));
                                break;
                        case 7:
                                // tuxgraphicsHWD.intvoltage0=1.3.6.1.4.1.ENTERPRISE_OID_NUM.7
                                // We need to byteswap the integer such
                                // that it is in network order:
                                volt_net_order=byteswap(volt_times_100[1]);
                                // send the 2 byte long number:
                                totlen_i_seq+=snmp_fill_oid_val_pair(snmpdat+offset_to_seq+2+totlen_i_seq,myoid,4,0x2,(uint8_t *)&volt_net_order,2); // we are passing 2 times 8 bit in volt_times_100
                                break;
                                // default is not needed 
                                // because of snmp_ckeck_if_all_are_my_oids
                }
                j++;
        }
       // write the length in the outer oid-seq tag 
       *(snmpdat+offset_to_seq+1)=totlen_i_seq;
       // calculate the udp len:
       snmpdatalen=offset_to_seq+totlen_i_seq+2;
       // set the length field of the oueter enclosing seq tag of the whole snmp message (the second udp byte):
       *(snmpdat+1)=offset_to_seq+totlen_i_seq;
       // set the request_type=0xa2 = snmp response (note: *(snmpdat+6) is the community string length)
       community_str_len=*(snmpdat+6);
       *(snmpdat+6+community_str_len+1)=0xa2;
       // set the len of the snmp response
       requIDlen=*(snmpdat+6+community_str_len+1+3);
       *(snmpdat+6+community_str_len+2)=totlen_i_seq+6+requIDlen+2+2; 
       return(snmpdatalen);
}

// end of functions for snmp 

int main(void){

        uint16_t pktlen;
        int8_t i,requestType;
        uint8_t snmpdatalen;
        uint8_t *snmpdat;
        uint8_t offset_to_seq;
        uint8_t snmp_num_of_oids;
        // do not answer more than 3 questions in one. Otherwise we
        // will have problems with the uint8_t datatype and the length:
#define MAX_NUM_OF_OID 3
        uint8_t last_digits_of_oids[MAX_NUM_OF_OID];
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        // LED
        DDRB|= (1<<DDB1); // enable PB1, LED as output
        LEDOFF;

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PIND7);// transistor off

        // read eeprom values unless, there are no valid values.
        if (eeprom_read_byte((uint8_t *)20) == 19 ){
                // ok magic number matches accept values
                eeprom_read_block((uint8_t *)gwip,(void *)21,sizeof(gwip));
                eeprom_read_block((uint8_t *)monitoredhost,(void *)26,sizeof(monitoredhost));
                pinginterval=eeprom_read_byte((uint8_t *)31);
                sendping=eeprom_read_byte((uint8_t *)32);
        }
        // accellerate the status update:
        if (sendping && pinginterval>5){
                pingtimer=pinginterval-4;
        }

        //initialize enc28j60:
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us

        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_loop_1(0); // 60us

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);
        if (gwip[0]==0){
                client_set_gwip(monitoredhost);
        }else{
                client_set_gwip(gwip);  // e.g internal IP of dsl router
        }
        timer_init();
        sei(); // interrupt enable
        adc_isr_on();

        // this is where we check if we got a ping from the monitored host:
        register_ping_rec_callback(&ping_callback);

        while(1){
                pktlen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                gPlen=packetloop_icmp_tcp(buf,pktlen);
                if (pktlen){
                        // get the next new packet from monitoredhost:
                        if(packetloop_icmp_checkreply(buf,monitoredhost)){
                                // ping reply is from monitored host and ping was from us
                                hoststatus=0;
                                hostbackafterreboot=1;
                                // host back we got a ping reply, LEDOFF:
                                //LEDOFF;
                                continue;
                        }
                }else{
                        // pktlen==0, nothing to do from network side
                        // spontanious messages must not interfer with
                        // web pages. We are idle here (idle task):
                        if (measurement_timer > 1){ 
                                measurement_timer=0;
                                volt_times_100[0]=adc_to_volttimes100(0);
                                volt_times_100[1]=adc_to_volttimes100(1);
                        }
                        if (sendping && client_waiting_gw()==0 && pingnow){
                                pingnow=0;
                                client_icmp_request(buf,monitoredhost);
                                pingtimer++; // otherweise we would call this function again
                                step_hoststatus();
                        }
                        if (do_extra_arp){
                                do_extra_arp=0;
                                if (gwip[0]==0){
                                        client_arp_whohas(buf,monitoredhost);
                                }else{
                                        client_arp_whohas(buf,gwip);  
                                }
                        }
                        if (reinitmac){
                                reinitmac=0;
                                if (gwip[0]==0){
                                        client_set_gwip(monitoredhost);
                                }else{
                                        client_set_gwip(gwip);  // e.g internal IP of dsl router
                                }
                        }
                        // end of idle task
                }
                if(gPlen==0){
                        // otherwise check for any other UDP messages.
                        goto UDP;
                }

                if (strncmp("GET ",(char *)&(buf[gPlen]),4)!=0){
                        // head, post and other methods:
                        gPlen=http200ok();
                        gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>200 OK</h1>"));
                        goto SENDTCP;
                }
                i=analyse_get_url((char *)&(buf[gPlen+5]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (i==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1><a href=/>continue</a>\n"));
                        goto SENDTCP;
                }
                if (i==0){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                        goto SENDTCP;
                }
                if (i==10){
                        goto SENDTCP;
                }
                // i==1 goes here
                gPlen=print_webpage_main(buf);
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send data
                continue;
                // tcp port www end
                // -------------------------------
                // udp start, we listen on snmp, port 161=0x00a1
UDP:
                if(eth_type_is_ip_and_my_ip(buf,pktlen)==0){
                        continue;
                }
                if (buf[IP_PROTO_P]==IP_PROTO_UDP_V&&buf[UDP_DST_PORT_L_P]==0xa1&&buf[UDP_DST_PORT_H_P]==0){
                        // the upper byte of the udp len should
                        // be zero as snmp get messages are short.
                        // We use only the lower byte:
                        snmpdatalen=buf[UDP_LEN_L_P]-UDP_HEADER_LEN;
                        snmpdat=&(buf[UDP_DATA_P]);
                        if (snmp_sanitycheck(snmpdat,snmpdatalen)==0){
                                // not a valid snmp message
                                continue;
                        }
                        if (snmp_check_community(snmpdat,my_snmp_community)==0){
                                // not our community (password), do not reply
                                continue;
                        }
                        requestType=snmp_check_get_or_getnext(snmpdat,&offset_to_seq);
                        if(requestType){ // get request or getnext
                                snmp_num_of_oids=snmp_numofSeq(snmpdat+offset_to_seq);
                                if (snmp_num_of_oids<1){
                                        continue; // nonsence message
                                }
                                if (snmp_num_of_oids>MAX_NUM_OF_OID){
                                        // answer tooBig:
                                        snmp_make_error_response(snmpdat,1,MAX_NUM_OF_OID+1);
                                        //snmpdatalen did not change
                                        goto UDPANSWER;
                                }
                                if (requestType==1){ // get
                                        i=snmp_ckeck_if_all_are_my_oids(snmpdat+offset_to_seq,snmp_num_of_oids,last_digits_of_oids);
                                        if (i){
                                                // no such OID (noSuchName)
                                                snmp_make_error_response(snmpdat,2,i);
                                                //snmpdatalen did not change
                                                goto UDPANSWER;
                                        }
                                }else{ // getnext
                                        i=snmp_find_oids_from_getnext(snmpdat+offset_to_seq,snmp_num_of_oids,last_digits_of_oids);
                                        if (i){
                                                // no such OID (noSuchName)
                                                snmp_make_error_response(snmpdat,2,i);
                                                //snmpdatalen did not change
                                                goto UDPANSWER;
                                        }
                                }
                                snmpdatalen=snmp_fill_answer_with_data(snmpdat,offset_to_seq,last_digits_of_oids,snmp_num_of_oids);
                               goto UDPANSWER;
                        }
                }
                continue;
UDPANSWER:
                make_udp_reply_from_request_udpdat_ready(buf,snmpdatalen,161);
        }
        return (0);
}
